#include <GL/glut.h>
#include <stdlib.h>
#include <math.h>
#include "polygonize.h"
#include "merge.h"
#include "patch.h"
#include "g:/school/bachelorscriptie/gl2ps-0.6/gl2ps.h"

#define ORTHOGRAPHIC
#undef ORTHOGRAPHIC
#define DRAWCUBES
#undef DRAWCUBES

#undef ERROR
#define ERROR -1

Vector3 eye(150,100,275);

int total, failed[5];

float time = 0;

extern "C"  {
	void prepare_f();
	double f(double x, double y, double z);
	int sizes[];
}

double f(Vector3 p)
{
	return f(p.x, p.y, p.z);
}

void callback(Vector3 a, Vector3 b, Vector3 c)
{
	// determine normal -- we only want to draw front-facing polygons (for the ps output to the thesis)
	Vector3 u(a.x-b.x, a.y-b.y, a.z-b.z), v(a.x-c.x, a.y-c.y, a.z-c.z);
	Vector3 n(u.y*v.z-u.z*v.y, u.z*v.x-u.x*v.z, u.x*v.y-u.y*v.x);
	Vector3 view(eye.x-a.x, eye.y-a.y, eye.z-a.z);
	float dotproduct = n.x*view.x + n.y*view.y + n.z*view.z;
	
	if (dotproduct > 0)
		glColor3f(0.5,0.5,0.5);
	else
		glColor3f(0,0,0);
	
	glBegin(GL_LINE_LOOP);
		glVertex3f(a.x, a.y, a.z);
		glVertex3f(b.x, b.y, b.z);
		glVertex3f(c.x, c.y, c.z);
	glEnd();
}

Octree buildOctree(Cube cube, int depth, Octree parent)
{
	if (depth >= maxDepth) 
		return NULL;

	int type = classify(&cube);

	if (depth < minDepth || type != 0)
	{
		Octree octree = new (struct pOctree);

		Vector3 min = cube.min, max = cube.max, mid;
		mid.x = (min.x + max.x) / 2;
		mid.y = (min.y + max.y) / 2;
		mid.z = (min.z + max.z) / 2;

		octree->cube = cube;
		octree->parent = parent;
		octree->children[0] = buildOctree(Cube(Vector3(min.x,min.y,max.z),mid,f), depth+1, octree);
		octree->children[1] = buildOctree(Cube(Vector3(max.x,min.y,max.z),mid,f), depth+1, octree);
		octree->children[2] = buildOctree(Cube(Vector3(max.x,min.y,min.z),mid,f), depth+1, octree);
		octree->children[3] = buildOctree(Cube(Vector3(min.x,min.y,min.z),mid,f), depth+1, octree);
		octree->children[4] = buildOctree(Cube(Vector3(min.x,max.y,max.z),mid,f), depth+1, octree);
		octree->children[5] = buildOctree(Cube(Vector3(max.x,max.y,max.z),mid,f), depth+1, octree);
		octree->children[6] = buildOctree(Cube(Vector3(max.x,max.y,min.z),mid,f), depth+1, octree);
		octree->children[7] = buildOctree(Cube(Vector3(min.x,max.y,min.z),mid,f), depth+1, octree);

		octree->leaf = octree->children[7] == NULL(octree->children[7] == NULL);
		octree->mergingFailed = false;
		calculateIntersections(&(octree->cube));

		return octree;
	}
	else
		return NULL;
}

void draw(Cube cube)
{
#ifdef DRAWCUBES
	Vector3 points[8];
	points[0] = Vector3(cube.min.x,cube.min.y,cube.min.z);
	points[1] = Vector3(cube.max.x,cube.min.y,cube.min.z);
	points[2] = Vector3(cube.max.x,cube.min.y,cube.max.z);
	points[3] = Vector3(cube.min.x,cube.min.y,cube.max.z);
	points[4] = Vector3(cube.min.x,cube.max.y,cube.min.z);
	points[5] = Vector3(cube.max.x,cube.max.y,cube.min.z);
	points[6] = Vector3(cube.max.x,cube.max.y,cube.max.z);
	points[7] = Vector3(cube.min.x,cube.max.y,cube.max.z);

   glColor3f(.5,.5,.5);
   glBegin(GL_LINES);
      glVertex3f(points[0].x, points[0].y, points[0].z);
      glVertex3f(points[1].x, points[1].y, points[1].z);

      glVertex3f(points[1].x, points[1].y, points[1].z);
      glVertex3f(points[2].x, points[2].y, points[2].z);

      glVertex3f(points[2].x, points[2].y, points[2].z);
      glVertex3f(points[3].x, points[3].y, points[3].z);

      glVertex3f(points[3].x, points[3].y, points[3].z);
      glVertex3f(points[0].x, points[0].y, points[0].z);

      glVertex3f(points[4].x, points[4].y, points[4].z);
      glVertex3f(points[5].x, points[5].y, points[5].z);

      glVertex3f(points[5].x, points[5].y, points[5].z);
      glVertex3f(points[6].x, points[6].y, points[6].z);

      glVertex3f(points[6].x, points[6].y, points[6].z);
      glVertex3f(points[7].x, points[7].y, points[7].z);

      glVertex3f(points[7].x, points[7].y, points[7].z);
      glVertex3f(points[4].x, points[4].y, points[4].z);

      glVertex3f(points[0].x, points[0].y, points[0].z);
      glVertex3f(points[4].x, points[4].y, points[4].z);

      glVertex3f(points[1].x, points[1].y, points[1].z);
      glVertex3f(points[5].x, points[5].y, points[5].z);

      glVertex3f(points[2].x, points[2].y, points[2].z);
      glVertex3f(points[6].x, points[6].y, points[6].z);

      glVertex3f(points[3].x, points[3].y, points[3].z);
      glVertex3f(points[7].x, points[7].y, points[7].z);

	glEnd();

#endif

	glColor3f(0,0,0);

	polygonize(cube, &callback);
}

void draw(Octree octree)
{
	if (octree == NULL)
		return;

	if (octree->leaf)
		draw(octree->cube);
	else
		for (int i=0; i<8; i++)
			draw(octree->children[i]);
}

void initialize() 
{
	initializePolygonisation();

	glClearColor(1,1,1,0) ;
 
	glMatrixMode(GL_PROJECTION); 
	glLoadIdentity();
#ifdef ORTHOGRAPHIC
	gluOrtho2D(0,4,0,4);
#else
	gluPerspective(45, 1, 1, 1024);
#endif

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
#ifndef ORTHOGRAPHIC
	gluLookAt( eye.x, eye.y, eye.z,  
			   sizes[0]/2, sizes[1]/2-100, sizes[2]/2,   
			   0,1,0 );
	glRotatef(45,190,0,0);
#endif
}

bool pause = false, outputPS = false;

void output(Octree octree)
{
	if (outputPS)
	{
		FILE *fp = fopen("g:/school/bachelorscriptie/test.ps", "w");
		int buffsize = 0, state = GL2PS_OVERFLOW;

		while( state == GL2PS_OVERFLOW ){ 
		  buffsize += 1024*1024;
		  gl2psBeginPage ( "MyTitle", "MySoftware", 
						   GL2PS_PS, GL2PS_BSP_SORT,
						   GL2PS_SIMPLE_LINE_OFFSET | GL2PS_SILENT, 
						   GL_RGBA, 0, NULL, buffsize, fp, NULL );
  		  gl2psLineWidth(0.5);
  		  draw(octree);
		  state = gl2psEndPage();
		}

		fclose(fp);

		outputPS = false;
	}
	else
		draw(octree);
}

void glDisplay() 
{
	if (pause) return;

	glClear(GL_COLOR_BUFFER_BIT) ;
#ifndef ORTHOGRAPHIC
//	glRotatef(0.1,0,1,1) ;
#endif

// the actual work goes here
	double size = 1.25;
	Octree octree = buildOctree(Cube(Vector3(0,0,0), 
		                             Vector3(sizes[0]-1, sizes[1]-1, sizes[2]-1), f), 0, NULL);

total = 0;
for (int i=0; i<5; i++)
	failed[i] = 0;

	merge(&octree, ERROR);

//	patch(octree);
	output(octree);
	
	kill(octree);
// and it is done

printf("%d\t%d\t%d\t%d\t%d\t%d\n", total, failed[0], failed[1], failed[2], failed[3], failed[4]);

	glutSwapBuffers() ;
}

void glIdle() 
{
	if (pause) return;
	time += 0.5f;
	glutPostRedisplay() ;
}

void glKey(unsigned char key, int x, int y)
{
	if (key == 'p')
		pause = !pause;
	else if (key == ' ')
		outputPS = true;
	else
		exit(0);
}

int main(int argc, char **argv) 
{
	prepare_f();
	glutInit( &argc, argv ) ;
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGBA);
	glutInitWindowSize(640,480);
	glutCreateWindow("Marching Cubes algorithm");
	glutDisplayFunc(glDisplay);
	glutKeyboardFunc(glKey);
	glutIdleFunc(glIdle);	
	
	initialize();
	
printf("total \t 1 \t 2 \t 3 \t 4 \t 5\n");

	glutMainLoop();

	return 0;
}



