#include <list>
#include <math.h>
#include "merge.h"
extern int total, failed[5];
using namespace std;

// for a succesful merge, all 8 child cells must either be "simple" cases or null
bool condition1(Octree octree)
{
	if (octree == NULL)
		return false;

	for (int i=0; i<8; i++)
	{
		Octree child = octree->children[i];
		if (child != NULL)
			if (!cases[child->cube.vertexFlags].simple)
 				return false;
	}

	return true;
}

// for a succesful merge, merging may not have failed at a lower level
bool condition2(Octree octree)
{
	if (octree == NULL)
		return true;

	if (octree->mergingFailed)
		return false;

	return condition2(octree->children[0]) &&
	       condition2(octree->children[1]) &&
	       condition2(octree->children[2]) &&
	       condition2(octree->children[3]) &&
	       condition2(octree->children[4]) &&
	       condition2(octree->children[5]) &&
	       condition2(octree->children[6]) &&
	       condition2(octree->children[7]);
}

// for a succesful merge, the resulting parent cell may not have more than one intersection 
// on any of its edges
bool condition3(Octree octree)
{
	int edgeFlags[8];
	for (int i=0; i<8; i++)
		if (octree->children[i] == NULL)
			edgeFlags[i] = 0;
		else
			edgeFlags[i] = cases[octree->children[i]->cube.vertexFlags].edgeFlags;

	// edge 0
	if (getBit(edgeFlags[0], 0) && getBit(edgeFlags[1], 0)) return false;

	// edge 1
	if (getBit(edgeFlags[1], 1) && getBit(edgeFlags[2], 1)) return false;

	// edge 2
	if (getBit(edgeFlags[2], 2) && getBit(edgeFlags[3], 2)) return false;

	// edge 3
	if (getBit(edgeFlags[3], 3) && getBit(edgeFlags[0], 3)) return false;

	// edge 4
	if (getBit(edgeFlags[4], 4) && getBit(edgeFlags[5], 4)) return false;

	// edge 5
	if (getBit(edgeFlags[5], 5) && getBit(edgeFlags[6], 5)) return false;

	// edge 6
	if (getBit(edgeFlags[6], 6) && getBit(edgeFlags[7], 6)) return false;

	// edge 7
	if (getBit(edgeFlags[7], 7) && getBit(edgeFlags[4], 7)) return false;

	// edge 8
	if (getBit(edgeFlags[0], 8) && getBit(edgeFlags[4], 8)) return false;

	// edge 9
	if (getBit(edgeFlags[1], 9) && getBit(edgeFlags[5], 9)) return false;

	// edge 10
	if (getBit(edgeFlags[2], 10) && getBit(edgeFlags[6], 10)) return false;

	// edge 11
	if (getBit(edgeFlags[3], 11) && getBit(edgeFlags[7], 11)) return false;

	// internal edge - x
	if (getBit(edgeFlags[0],  6) && getBit(edgeFlags[1],  6) &&
		getBit(edgeFlags[2],  4) && getBit(edgeFlags[3],  4) &&
		getBit(edgeFlags[4],  2) && getBit(edgeFlags[5],  2) &&
		getBit(edgeFlags[6],  0) && getBit(edgeFlags[7],  0) ) return false;

	// internal edge - y
	if (getBit(edgeFlags[0], 10) && getBit(edgeFlags[1], 11) &&
		getBit(edgeFlags[2],  8) && getBit(edgeFlags[3],  9) &&
		getBit(edgeFlags[4], 10) && getBit(edgeFlags[5], 11) &&
		getBit(edgeFlags[6],  8) && getBit(edgeFlags[7],  9) ) return false;
	
	// internal edge - z
	if (getBit(edgeFlags[0],  5) && getBit(edgeFlags[1],  7) &&
		getBit(edgeFlags[2],  7) && getBit(edgeFlags[3],  5) &&
		getBit(edgeFlags[4],  1) && getBit(edgeFlags[5],  3) &&
		getBit(edgeFlags[6],  3) && getBit(edgeFlags[7],  1) ) return false;

	return true;
}


void combineEdges(Octree octree, int edge, int cube1, int cube2, int edgeFlag, int edgeFlags[])
{
	if (getBit(edgeFlag, edge))
	{
		if (getBit(edgeFlags[cube1], edge))
			octree->cube.intersections[edge] = octree->children[cube1]->cube.intersections[edge];
		else if (getBit(edgeFlags[cube2], edge))
			octree->cube.intersections[edge] = octree->children[cube2]->cube.intersections[edge] ;
	}
}

// combine intersection points of the child cells to obtain intersection points for the
// resulting parent cell
void combineIntersections(Octree octree)
{
	octree->cube.vertexFlags = makeVertexFlags(
		octree->children[0] == NULL ? 0 : getBit(octree->children[0]->cube.vertexFlags, 0),
		octree->children[1] == NULL ? 0 : getBit(octree->children[1]->cube.vertexFlags, 1),
		octree->children[2] == NULL ? 0 : getBit(octree->children[2]->cube.vertexFlags, 2),
		octree->children[3] == NULL ? 0 : getBit(octree->children[3]->cube.vertexFlags, 3),
		octree->children[4] == NULL ? 0 : getBit(octree->children[4]->cube.vertexFlags, 4),
		octree->children[5] == NULL ? 0 : getBit(octree->children[5]->cube.vertexFlags, 5),
		octree->children[6] == NULL ? 0 : getBit(octree->children[6]->cube.vertexFlags, 6),
		octree->children[7] == NULL ? 0 : getBit(octree->children[7]->cube.vertexFlags, 7)
		);
	int edgeFlag = cases[octree->cube.vertexFlags].edgeFlags;

	int edgeFlags[8];
	for (int i=0; i<8; i++)
		if (octree->children[i] == NULL)
			edgeFlags[i] = (1 << 13);
		else
			edgeFlags[i] = cases[octree->children[i]->cube.vertexFlags].edgeFlags;

	combineEdges(octree, 0, 0, 1, edgeFlag, edgeFlags);
	combineEdges(octree, 1, 1, 2, edgeFlag, edgeFlags);
	combineEdges(octree, 2, 2, 3, edgeFlag, edgeFlags);
	combineEdges(octree, 3, 3, 0, edgeFlag, edgeFlags);
	combineEdges(octree, 4, 4, 5, edgeFlag, edgeFlags);
	combineEdges(octree, 5, 5, 6, edgeFlag, edgeFlags);
	combineEdges(octree, 6, 6, 7, edgeFlag, edgeFlags);
	combineEdges(octree, 7, 7, 4, edgeFlag, edgeFlags);
	combineEdges(octree, 8, 0, 4, edgeFlag, edgeFlags);
	combineEdges(octree, 9, 1, 5, edgeFlag, edgeFlags);
	combineEdges(octree, 10, 2, 6, edgeFlag, edgeFlags);
	combineEdges(octree, 11, 3, 7, edgeFlag, edgeFlags);
}


// for a succesful merge, the resulting parent cell must have a "simple" case
// (otherwise we will get trouble with crack patching later)
bool condition4(Octree octree)
{
	classify(&(octree->cube));

	return cases[octree->cube.vertexFlags].simple;
}


/*
   Solve a system of 3 equations in 3 unknowns using Gaussian Elimination
   Solve an equation in matrix form Ax = 0
   The 2D array a is the matrix A:

         /  a[0][0]  a[0][1]  a[0][2]  \
	 A = |  a[1][0]  a[1][1]  a[1][2]  |
	     \  a[2][0]  a[2][1]  a[2][2]  /

   The result is returned in x, otherwise the function returns false
   if the system of equations is singular.
*/
#define ABS(a) ((a)<0 ? -(a) : (a))

bool GaussianSolve(float a[4][4], float *x)
{
   int i,j,k,maxrow;

   for (i=0;i<4;i++) 
   {
      // Find the row with the largest first value 
      maxrow = i;
      for (j=i+1;j<4;j++) 
         if (ABS(a[j][i]) > ABS(a[maxrow][i]))
            maxrow = j;

      // Swap the maxrow and ith row 
      for (k=0;k<4;k++) 
	  {
         float tmp = a[i][k];
         a[i][k] = a[maxrow][k];
         a[maxrow][k] = tmp;
      }

      // Singular matrix? 
      if (ABS(a[i][i]) < 0.0001)
		return false;

	  // Eliminate the ith element of the jth row 
	  for (j=i+1;j<4;j++) 
		 for (k=4;k>=i;k--) 
			a[j][k] -= a[i][k] * a[j][i] / a[i][i];
   }

   // Do the back substitution
   for (j=3;j>=0;j--) 
   {
      float tmp = 0;
      for (k=j+1;k<4;k++)
         tmp += a[j][k] * x[k];
      x[j] = -tmp / a[j][j];
   }

   return true;
}

float distance(Vector3 p, float A, float B, float C, float D)
{
	return A*p.x + B*p.y + C*p.z + D;
}

float maxDistance(Octree octree, float A, float B, float C, float D, float soFar)
{
	if (octree == NULL)
		return soFar;

	bool leaf = true;
	for (int i=0; i<8; i++)
		if (octree->children[i] != NULL)
			leaf = false;

	if (leaf)
	{
		int edgeFlags = cases[octree->cube.vertexFlags].edgeFlags;
		for (i=0; i<12; i++)
			if (getBit(edgeFlags, i))
			{
				float d = distance(octree->cube.intersections[i], A,B,C,D);
				if (d > soFar)
					soFar = d;
			}
	}
	else
		for (i=0; i<8; i++)
			soFar = maxDistance(octree->children[i], A,B,C,D, soFar);

	return soFar;
}

float buildEntry(float x[12], float xc, float y[12], float yc, int n)
{
	float result = 0;
	for (int i=0; i<n; i++)
		result += (x[i] - xc) * (y[i] - yc);

	return result;
}

// test if the resulting (merged) parent cell will be within the user-specified
// error range
bool condition5(Octree octree, float error)
{
	// gather the intersection points in a vector
	int edgeFlags = cases[octree->cube.vertexFlags].edgeFlags;
	int n = 0;
	float x[12], y[12], z[12], xc=0, yc=0, zc=0;
	for (int i=0; i<12; i++)
		if (getBit(edgeFlags, i))
		{
			x[n] = octree->cube.intersections[i].x;		xc += x[n];
			y[n] = octree->cube.intersections[i].y;		yc += y[n];
			z[n] = octree->cube.intersections[i].z;		zc += z[n];
			n++;
		}

	xc /= n;
	yc /= n;
	zc /= n;

	// build the matrix to solve for the least-squares-plane
	float a[4][4];
/*	a[0][0] = buildEntry(x, xc, x, xc, n);
	a[1][0] = buildEntry(y, yc, x, xc, n);
	a[2][0] = buildEntry(z, zc, x, xc, n);

	a[0][1] = buildEntry(x, xc, y, yc, n);
	a[1][1] = buildEntry(y, yc, y, yc, n);
	a[2][1] = buildEntry(z, zc, y, yc, n);

	a[0][2] = buildEntry(x, xc, z, zc, n);
	a[1][2] = buildEntry(y, yc, z, zc, n);
	a[2][2] = buildEntry(z, zc, z, zc, n);
*/
	a[0][0] = buildEntry(x, 0, x, 0, n);
	a[1][0] = buildEntry(y, 0, x, 0, n);
	a[2][0] = buildEntry(z, 0, x, 0, n);
	a[3][0] = - n * xc;

	a[0][1] = buildEntry(x, 0, y, 0, n);
	a[1][1] = buildEntry(y, 0, y, 0, n);
	a[2][1] = buildEntry(z, 0, y, 0, n);
	a[3][1] = - n * yc;

	a[0][2] = buildEntry(x, 0, z, 0, n);
	a[1][2] = buildEntry(y, 0, z, 0, n);
	a[2][2] = buildEntry(z, 0, z, 0, n);
	a[3][2] = - n * zc;

	a[0][3] = n * xc;
	a[1][3] = n * yc;
	a[2][3] = n * zc;
	a[3][3] = -1;

	// solve it to find the least-squares-plane
	float plane[4];
	if (!GaussianSolve(a,plane))
		return false;

	// normalize the plane
	float length = sqrt(plane[0]*plane[0] + plane[1]*plane[1] + plane[2]*plane[2]);
	float A = plane[0] / length;
	float B = plane[1] / length;
	float C = plane[2] / length;
	float D = plane[3];//A*xc+B*yc+C*zc;

	double d = maxDistance(octree, A, B, C, D, 0);
 	return d < error;
}

void fail(Octree octree)
{
	if (octree == NULL) return;

	octree->mergingFailed = true;
	
	if (octree->leaf) 
		for (int i=0; i<8; i++)
		{
			kill(octree->children[i]);
			octree->children[i] = NULL;
		}
	else
		for (int i=0; i<8; i++)
			fail(octree->children[i]);
}

void doMerge(Octree *octree, float error)
{
	(*octree)->mergingFailed = false;

	// work bottom-up through the octree, starting at the lowest non-leaf layer
	bool leaf = true;
	for (int i=0; i<8; i++)
		if ((*octree)->children[i] != NULL) leaf = false;

	if (leaf)
	{
		(*octree)->leaf = true;
		calculateIntersections(&((*octree)->cube));
	}
	else
	{
		(*octree)->leaf = false;

		// check if the conditions apply
		if (condition1(*octree))
			if (condition2(*octree))
				if (condition3(*octree))
					if (condition4(*octree))
					{
//						combineIntersections(*octree);
						calculateIntersections(&((*octree)->cube));
						if (condition5(*octree, error))
						{
							// if so, we have a merge
							total++;
							(*octree)->leaf = true;
							(*octree)->mergingFailed = false;
						} else { fail(*octree); failed[4]++; }
					}
					else { fail(*octree); failed[3]++; }
				else { fail(*octree); failed[2]++; }
			else { fail(*octree); failed[1]++; }
		else { fail(*octree); failed[0]++; }
	}
}

void Traverse(list<Octree *> l, float error)
{
	if (l.empty())
		return;

	list<Octree *> children;
	for (list<Octree *>::iterator i=l.begin(); i!=l.end(); i++)
	{
		Octree *o = *i;
		if (o != NULL && *o != NULL)
			for (int j=0; j<8; j++)
				if ((*o)->children[j] != NULL)
					children.push_back(&((*o)->children[j]));
	}

	Traverse(children, error);

	for (i=l.begin(); i!=l.end(); i++)
	{
		Octree *o = *i;
		if (o != NULL && *o != NULL)
			doMerge(o, error);
	}
}

void merge(Octree *octree, float error)
{
/*	if (*octree == NULL)
		return;

	list<Octree *> q;
	q.push_back(octree);

	// traverse the octree in a breadth-first manner
	while (!q.empty())
	{
		octree = q.front();
		q.pop_front();

		for (int j=0; j<8; j++)
			if ((*octree)->children[j] != NULL)
				q.push_back(&((*octree)->children[j]));

		doMerge(octree, error);
	}*/

	list<Octree *> l;
	l.push_front(octree);
	Traverse(l, error);
}

void kill(Octree octree)
{
	if (octree == NULL)
		return;

	for (int i=0; i<8; i++)
		kill(octree->children[i]);

	delete octree;
}

