#include "merge.h"
#include "patch.h"
#include <queue>

using namespace std;

typedef enum DIRECTION { LEFT, RIGHT, TOP, BOTTOM, FAR, NEAR };

Octree getNeighbour(Octree cell, DIRECTION neighbour);

Octree getSmallCell(Octree parent, int side, DIRECTION neighbour)
{
	if (side<10)
		return parent->children[side];
	else {
		Octree o = getNeighbour(parent, neighbour);
		return o == NULL ? o : o->children[side-10];
	}
}

Octree getSmallCell(DIRECTION neighbour, int Top, int Right, int Near, Octree parent)
{
	int Bottom	= Top<10	? Top+10	: Top-10;
	int Left	= Right<10	? Right+10	: Right-10;
	int Far		= Near<10	? Near+10	: Near-10;

	switch(neighbour)
	{
		case TOP:		return getSmallCell(parent, Top,	TOP);
		case BOTTOM:	return getSmallCell(parent, Bottom,	BOTTOM);
		case LEFT:		return getSmallCell(parent, Left,	LEFT);
		case RIGHT:		return getSmallCell(parent, Right,	RIGHT);
		case NEAR:		return getSmallCell(parent, Near,	NEAR);
		case FAR:		return getSmallCell(parent, Far,	FAR);
	}
	return NULL;
}

Octree getNeighbour(Octree cell, DIRECTION neighbour)
{
	Octree parent = cell->parent;
	if (parent == NULL) return NULL;

	int position = -1;
	for (int pos = 0; pos<8; pos++)
		if (parent->children[pos] == cell)
			position = pos;

	switch(position)
	{
		case 0: return getSmallCell(neighbour,  4, 1, 3,parent);
		case 1:	return getSmallCell(neighbour,  5,10, 2,parent);
		case 2:	return getSmallCell(neighbour,  6,13,11,parent);
		case 3:	return getSmallCell(neighbour,  7, 2,10,parent);
		case 4:	return getSmallCell(neighbour, 10, 5, 7,parent);
		case 5:	return getSmallCell(neighbour, 11,14, 6,parent);
		case 6:	return getSmallCell(neighbour, 12,17,15,parent);
		case 7:	return getSmallCell(neighbour, 13, 6,14,parent);
	}

	return NULL;
}

#define ABS(x) ((x)<0 ? -(x) : (x))

void collapse(Vector3 &x, Vector3 a, Vector3 b)
{
	double l = (a.x-b.x)*(a.x-b.x) + (a.y-b.y)*(a.y-b.y) + (a.z-b.z)*(a.z-b.z);
	if (ABS(l)<0.0000001) return;
	l = ((x.x-b.x)*(a.x-b.x) + (x.y-b.y)*(a.y-b.y) + (x.z-b.z)*(a.z-b.z)) / l;

	x.x = b.x + l*(a.x-b.x);
	x.y = b.y + l*(a.y-b.y);
	x.z = b.z + l*(a.z-b.z);
}

void collapse(Octree cell, int p1, int p2, int p3, int p4, Vector3 a, Vector3 b)
{
	int edgeFlags = cases[cell->cube.vertexFlags].edgeFlags;
	if (getBit(edgeFlags, p1)) collapse(cell->cube.intersections[p1], a, b);
	if (getBit(edgeFlags, p2)) collapse(cell->cube.intersections[p2], a, b);
	if (getBit(edgeFlags, p3)) collapse(cell->cube.intersections[p3], a, b);
	if (getBit(edgeFlags, p4)) collapse(cell->cube.intersections[p4], a, b);
}

void getLine(Octree cell, int p1, int p2, int p3, int p4, Vector3 &a, Vector3 &b)
{
	int edgeFlags = cases[cell->cube.vertexFlags].edgeFlags;
	bool f = true;
	if (getBit(edgeFlags, p1))
		if (f) { a = cell->cube.intersections[p1]; f = !f; }
		else     b = cell->cube.intersections[p1];
	else if (getBit(edgeFlags, p2))
		if (f) { a = cell->cube.intersections[p2]; f = !f; }
		else     b = cell->cube.intersections[p2];
	else if (getBit(edgeFlags, p3))
		if (f) { a = cell->cube.intersections[p3]; f = !f; }
		else     b = cell->cube.intersections[p3];
	else if (getBit(edgeFlags, p4))
		if (f) { a = cell->cube.intersections[p4]; f = !f; }
		else     b = cell->cube.intersections[p4];
}

void patchFace(Octree bigCell, DIRECTION bigFace, Octree smallCell, DIRECTION smallFace)
{
	if (smallCell == NULL) return;
	if (cases[smallCell->cube.vertexFlags].type == 0) return;

	if (!smallCell->leaf)
		for (int i=0; i<8; i++)
			patchFace(bigCell, bigFace, smallCell->children[i], smallFace);

	// determine the line on the low-resolution face, to which to collapse
	Vector3 a(0,0,0),b(0,0,0);
	switch(bigFace)
	{
		case LEFT:    getLine(bigCell, 3,8,7,11, a, b); break;
		case RIGHT:   getLine(bigCell, 1,9,5,10, a, b); break;
		case TOP:     getLine(bigCell, 4,5,6,7, a, b); break;
		case BOTTOM:  getLine(bigCell, 0,1,2,3, a, b); break;
		case NEAR:    getLine(bigCell, 2,10,6,11, a, b); break;
		case FAR:     getLine(bigCell, 0,8,4,9, a, b); break;
	}

	// and then collapse the high-resolution face to that line
	switch(smallFace)
	{
		case LEFT:    collapse(smallCell, 3,8,7,11, a, b); break;
		case RIGHT:   collapse(smallCell, 1,9,5,10, a, b); break;
		case TOP:     collapse(smallCell, 4,5,6,7, a, b); break;
		case BOTTOM:  collapse(smallCell, 0,1,2,3, a, b); break;
		case NEAR:    collapse(smallCell, 2,10,6,11, a, b); break;
		case FAR:     collapse(smallCell, 0,8,4,9, a, b); break;
	}
}

void patchSmallCells(Octree bigCell, DIRECTION bigFace, Octree smallCells, DIRECTION smallFace, int o1, int o2, int o3, int o4)
{
	patchFace(bigCell, bigFace, smallCells->children[o1], smallFace);
	patchFace(bigCell, bigFace, smallCells->children[o2], smallFace);
	patchFace(bigCell, bigFace, smallCells->children[o3], smallFace);
	patchFace(bigCell, bigFace, smallCells->children[o4], smallFace);
}

void patchInterface(Octree bigCell, DIRECTION neighbour)
{
	Octree smallCells = getNeighbour(bigCell, neighbour);

	if (smallCells == NULL) return;
	if (cases[smallCells->cube.vertexFlags].type == 0) return;

	switch(neighbour)
	{
	case TOP:	patchSmallCells(bigCell, TOP, smallCells, BOTTOM, 0,1,2,3); break;
	case BOTTOM:patchSmallCells(bigCell, BOTTOM, smallCells, TOP, 4,5,6,7); break;
	case LEFT:	patchSmallCells(bigCell, LEFT, smallCells, RIGHT, 1,2,5,6); break;
	case RIGHT:	patchSmallCells(bigCell, RIGHT, smallCells, LEFT, 0,3,7,4); break;
	case NEAR:  patchSmallCells(bigCell, NEAR, smallCells, FAR, 0,1,5,4); break;
	case FAR:	patchSmallCells(bigCell, FAR, smallCells, NEAR, 2,3,7,6); break;
	}
}

void patch(Octree octree)
{
	queue<Octree> q;
	q.push(octree);

	// traverse the octree in a breadth-first manner
	while (!q.empty())
	{
		Octree o = q.front();
		q.pop();

		if (cases[o->cube.vertexFlags].type != 0)
		{
			// do something with each cell (o):
			//		find its 6 connected neighbours on the same level
			//		if one of the neighbours is not a leaf,
			//		patch the interface on the child-cells
			patchInterface(o, LEFT);
			patchInterface(o, RIGHT);
			patchInterface(o, TOP);
			patchInterface(o, BOTTOM);
			patchInterface(o, FAR);
			patchInterface(o, NEAR);

			for (int i=0; i<8; i++)
				if (o->children[i] != NULL)
					q.push(o->children[i]);
		}
	}
}