#include <GL/glut.h>
#include <stdio.h>
#include "polygonize.h"

//#define NOTETRAHEDRA
//#undef NOTETRAHEDRA


Vector3 interpolate(Vector3 p1, Vector3 p2, double value1, double value2)
{
	double mu = (-value1) / (value2 - value1);
	Vector3 p;
	p.x = p1.x + mu * (p2.x - p1.x);
	p.y = p1.y + mu * (p2.y - p1.y);
	p.z = p1.z + mu * (p2.z - p1.z);
	return p;
}

// determine the type of the cube
// return whether the surface intersects this cube or not
int classify(Cube *cube)
{
   cube->vertexFlags = 0;
   for (int i=0; i<8; i++)
	   if (cube->value[i] < 0) 
		   cube->vertexFlags |= (1 << i);

   return cases[cube->vertexFlags].type;
}


void calculateIntersections(Cube *cube)
{
   // Determine the index into the edge table which 
   // tells us which vertices are inside of the surface 
   int vertexFlags = 0;
   for (int i=0; i<8; i++)
	   if (cube->value[i] < 0) 
		   vertexFlags |= (1 << i);

   // Cube is entirely in/out of the surface 
   if (cases[vertexFlags].type == 0)
      return;

   // Find the vertices where the surface intersects the cube 
   Vector3 points[8];
	points[0] = Vector3(cube->min.x,cube->min.y,cube->min.z);
	points[1] = Vector3(cube->max.x,cube->min.y,cube->min.z);
	points[2] = Vector3(cube->max.x,cube->min.y,cube->max.z);
	points[3] = Vector3(cube->min.x,cube->min.y,cube->max.z);
	points[4] = Vector3(cube->min.x,cube->max.y,cube->min.z);
	points[5] = Vector3(cube->max.x,cube->max.y,cube->min.z);
	points[6] = Vector3(cube->max.x,cube->max.y,cube->max.z);
	points[7] = Vector3(cube->min.x,cube->max.y,cube->max.z);

   if (getBit(cases[vertexFlags].edgeFlags, 0))
      cube->intersections[0] = interpolate(points[0],points[1],cube->value[0],cube->value[1]);
   if (getBit(cases[vertexFlags].edgeFlags, 1))
      cube->intersections[1] = interpolate(points[1],points[2],cube->value[1],cube->value[2]);
   if (getBit(cases[vertexFlags].edgeFlags, 2))
      cube->intersections[2] = interpolate(points[2],points[3],cube->value[2],cube->value[3]);
   if (getBit(cases[vertexFlags].edgeFlags, 3))
      cube->intersections[3] = interpolate(points[3],points[0],cube->value[3],cube->value[0]);
   if (getBit(cases[vertexFlags].edgeFlags, 4))
      cube->intersections[4] = interpolate(points[4],points[5],cube->value[4],cube->value[5]);
   if (getBit(cases[vertexFlags].edgeFlags, 5))
      cube->intersections[5] = interpolate(points[5],points[6],cube->value[5],cube->value[6]);
   if (getBit(cases[vertexFlags].edgeFlags, 6))
      cube->intersections[6] = interpolate(points[6],points[7],cube->value[6],cube->value[7]);
   if (getBit(cases[vertexFlags].edgeFlags, 7))
      cube->intersections[7] = interpolate(points[7],points[4],cube->value[7],cube->value[4]);
   if (getBit(cases[vertexFlags].edgeFlags, 8))
      cube->intersections[8] = interpolate(points[0],points[4],cube->value[0],cube->value[4]);
   if (getBit(cases[vertexFlags].edgeFlags, 9))
      cube->intersections[9] = interpolate(points[1],points[5],cube->value[1],cube->value[5]);
   if (getBit(cases[vertexFlags].edgeFlags, 10))
      cube->intersections[10] = interpolate(points[2],points[6],cube->value[2],cube->value[6]);
   if (getBit(cases[vertexFlags].edgeFlags, 11))
      cube->intersections[11] = interpolate(points[3],points[7],cube->value[3],cube->value[7]);
}

//#ifdef NOTETRAHEDRA

// which triangles to output on a given case
int triTable[256][16] =
{{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 8, 3, 9, 8, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 2, 10, 0, 2, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 8, 3, 2, 10, 8, 10, 9, 8, -1, -1, -1, -1, -1, -1, -1},
{3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 11, 2, 8, 11, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 11, 2, 1, 9, 11, 9, 8, 11, -1, -1, -1, -1, -1, -1, -1},
{3, 10, 1, 11, 10, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 10, 1, 0, 8, 10, 8, 11, 10, -1, -1, -1, -1, -1, -1, -1},
{3, 9, 0, 3, 11, 9, 11, 10, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 3, 0, 7, 3, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 1, 9, 4, 7, 1, 7, 3, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 4, 7, 3, 0, 4, 1, 2, 10, -1, -1, -1, -1, -1, -1, -1},
{9, 2, 10, 9, 0, 2, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
{2, 10, 9, 2, 9, 7, 2, 7, 3, 7, 9, 4, -1, -1, -1, -1},
{8, 4, 7, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 4, 7, 11, 2, 4, 2, 0, 4, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 1, 8, 4, 7, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
{4, 7, 11, 9, 4, 11, 9, 11, 2, 9, 2, 1, -1, -1, -1, -1},
{3, 10, 1, 3, 11, 10, 7, 8, 4, -1, -1, -1, -1, -1, -1, -1},
{1, 11, 10, 1, 4, 11, 1, 0, 4, 7, 11, 4, -1, -1, -1, -1},
{4, 7, 8, 9, 0, 11, 9, 11, 10, 11, 0, 3, -1, -1, -1, -1},
{4, 7, 11, 4, 11, 9, 9, 11, 10, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 5, 4, 1, 5, 0, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 5, 4, 8, 3, 5, 3, 1, 5, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 1, 2, 10, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
{5, 2, 10, 5, 4, 2, 4, 0, 2, -1, -1, -1, -1, -1, -1, -1},
{2, 10, 5, 3, 2, 5, 3, 5, 4, 3, 4, 8, -1, -1, -1, -1},
{9, 5, 4, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 11, 2, 0, 8, 11, 4, 9, 5, -1, -1, -1, -1, -1, -1, -1},
{0, 5, 4, 0, 1, 5, 2, 3, 11, -1, -1, -1, -1, -1, -1, -1},
{2, 1, 5, 2, 5, 8, 2, 8, 11, 4, 8, 5, -1, -1, -1, -1},
{10, 3, 11, 10, 1, 3, 9, 5, 4, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 5, 0, 8, 1, 8, 10, 1, 8, 11, 10, -1, -1, -1, -1},
{5, 4, 0, 5, 0, 11, 5, 11, 10, 11, 0, 3, -1, -1, -1, -1},
{5, 4, 8, 5, 8, 10, 10, 8, 11, -1, -1, -1, -1, -1, -1, -1},
{9, 7, 8, 5, 7, 9, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 3, 0, 9, 5, 3, 5, 7, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 7, 8, 0, 1, 7, 1, 5, 7, -1, -1, -1, -1, -1, -1, -1},
{1, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 7, 8, 9, 5, 7, 10, 1, 2, -1, -1, -1, -1, -1, -1, -1},
{10, 1, 2, 9, 5, 0, 5, 3, 0, 5, 7, 3, -1, -1, -1, -1},
{8, 0, 2, 8, 2, 5, 8, 5, 7, 10, 5, 2, -1, -1, -1, -1},
{2, 10, 5, 2, 5, 3, 3, 5, 7, -1, -1, -1, -1, -1, -1, -1},
{7, 9, 5, 7, 8, 9, 3, 11, 2, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 7, 9, 7, 2, 9, 2, 0, 2, 7, 11, -1, -1, -1, -1},
{2, 3, 11, 0, 1, 8, 1, 7, 8, 1, 5, 7, -1, -1, -1, -1},
{11, 2, 1, 11, 1, 7, 7, 1, 5, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 8, 8, 5, 7, 10, 1, 3, 10, 3, 11, -1, -1, -1, -1},
{5, 7, 0, 5, 0, 9, 7, 11, 0, 1, 0, 10, 11, 10, 0, -1},
{11, 10, 0, 11, 0, 3, 10, 5, 0, 8, 0, 7, 5, 7, 0, -1},
{11, 10, 5, 7, 11, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 1, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 8, 3, 1, 9, 8, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 5, 2, 6, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 5, 1, 2, 6, 3, 0, 8, -1, -1, -1, -1, -1, -1, -1},
{9, 6, 5, 9, 0, 6, 0, 2, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 9, 8, 5, 8, 2, 5, 2, 6, 3, 2, 8, -1, -1, -1, -1},
{2, 3, 11, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 0, 8, 11, 2, 0, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 1, 9, 2, 9, 11, 2, 9, 8, 11, -1, -1, -1, -1},
{6, 3, 11, 6, 5, 3, 5, 1, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 11, 0, 11, 5, 0, 5, 1, 5, 11, 6, -1, -1, -1, -1},
{3, 11, 6, 0, 3, 6, 0, 6, 5, 0, 5, 9, -1, -1, -1, -1},
{6, 5, 9, 6, 9, 11, 11, 9, 8, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 3, 0, 4, 7, 3, 6, 5, 10, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 5, 10, 6, 8, 4, 7, -1, -1, -1, -1, -1, -1, -1},
{10, 6, 5, 1, 9, 7, 1, 7, 3, 7, 9, 4, -1, -1, -1, -1},
{6, 1, 2, 6, 5, 1, 4, 7, 8, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 5, 5, 2, 6, 3, 0, 4, 3, 4, 7, -1, -1, -1, -1},
{8, 4, 7, 9, 0, 5, 0, 6, 5, 0, 2, 6, -1, -1, -1, -1},
{7, 3, 9, 7, 9, 4, 3, 2, 9, 5, 9, 6, 2, 6, 9, -1},
{3, 11, 2, 7, 8, 4, 10, 6, 5, -1, -1, -1, -1, -1, -1, -1},
{5, 10, 6, 4, 7, 2, 4, 2, 0, 2, 7, 11, -1, -1, -1, -1},
{0, 1, 9, 4, 7, 8, 2, 3, 11, 5, 10, 6, -1, -1, -1, -1},
{9, 2, 1, 9, 11, 2, 9, 4, 11, 7, 11, 4, 5, 10, 6, -1},
{8, 4, 7, 3, 11, 5, 3, 5, 1, 5, 11, 6, -1, -1, -1, -1},
{5, 1, 11, 5, 11, 6, 1, 0, 11, 7, 11, 4, 0, 4, 11, -1},
{0, 5, 9, 0, 6, 5, 0, 3, 6, 11, 6, 3, 8, 4, 7, -1},
{6, 5, 9, 6, 9, 11, 4, 7, 9, 7, 11, 9, -1, -1, -1, -1},
{10, 4, 9, 6, 4, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 10, 6, 4, 9, 10, 0, 8, 3, -1, -1, -1, -1, -1, -1, -1},
{10, 0, 1, 10, 6, 0, 6, 4, 0, -1, -1, -1, -1, -1, -1, -1},
{8, 3, 1, 8, 1, 6, 8, 6, 4, 6, 1, 10, -1, -1, -1, -1},
{1, 4, 9, 1, 2, 4, 2, 6, 4, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 1, 2, 9, 2, 4, 9, 2, 6, 4, -1, -1, -1, -1},
{0, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 3, 2, 8, 2, 4, 4, 2, 6, -1, -1, -1, -1, -1, -1, -1},
{10, 4, 9, 10, 6, 4, 11, 2, 3, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 2, 2, 8, 11, 4, 9, 10, 4, 10, 6, -1, -1, -1, -1},
{3, 11, 2, 0, 1, 6, 0, 6, 4, 6, 1, 10, -1, -1, -1, -1},
{6, 4, 1, 6, 1, 10, 4, 8, 1, 2, 1, 11, 8, 11, 1, -1},
{9, 6, 4, 9, 3, 6, 9, 1, 3, 11, 6, 3, -1, -1, -1, -1},
{8, 11, 1, 8, 1, 0, 11, 6, 1, 9, 1, 4, 6, 4, 1, -1},
{3, 11, 6, 3, 6, 0, 0, 6, 4, -1, -1, -1, -1, -1, -1, -1},
{6, 4, 8, 11, 6, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 10, 6, 7, 8, 10, 8, 9, 10, -1, -1, -1, -1, -1, -1, -1},
{0, 7, 3, 0, 10, 7, 0, 9, 10, 6, 7, 10, -1, -1, -1, -1},
{10, 6, 7, 1, 10, 7, 1, 7, 8, 1, 8, 0, -1, -1, -1, -1},
{10, 6, 7, 10, 7, 1, 1, 7, 3, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 6, 1, 6, 8, 1, 8, 9, 8, 6, 7, -1, -1, -1, -1},
{2, 6, 9, 2, 9, 1, 6, 7, 9, 0, 9, 3, 7, 3, 9, -1},
{7, 8, 0, 7, 0, 6, 6, 0, 2, -1, -1, -1, -1, -1, -1, -1},
{7, 3, 2, 6, 7, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 11, 10, 6, 8, 10, 8, 9, 8, 6, 7, -1, -1, -1, -1},
{2, 0, 7, 2, 7, 11, 0, 9, 7, 6, 7, 10, 9, 10, 7, -1},
{1, 8, 0, 1, 7, 8, 1, 10, 7, 6, 7, 10, 2, 3, 11, -1},
{11, 2, 1, 11, 1, 7, 10, 6, 1, 6, 7, 1, -1, -1, -1, -1},
{8, 9, 6, 8, 6, 7, 9, 1, 6, 11, 6, 3, 1, 3, 6, -1},
{0, 9, 1, 11, 6, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 8, 0, 7, 0, 6, 3, 11, 0, 11, 6, 0, -1, -1, -1, -1},
{7, 11, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 8, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 9, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 1, 9, 8, 3, 1, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
{10, 1, 2, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 3, 0, 8, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
{2, 9, 0, 2, 10, 9, 6, 11, 7, -1, -1, -1, -1, -1, -1, -1},
{6, 11, 7, 2, 10, 3, 10, 8, 3, 10, 9, 8, -1, -1, -1, -1},
{7, 2, 3, 6, 2, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{7, 0, 8, 7, 6, 0, 6, 2, 0, -1, -1, -1, -1, -1, -1, -1},
{2, 7, 6, 2, 3, 7, 0, 1, 9, -1, -1, -1, -1, -1, -1, -1},
{1, 6, 2, 1, 8, 6, 1, 9, 8, 8, 7, 6, -1, -1, -1, -1},
{10, 7, 6, 10, 1, 7, 1, 3, 7, -1, -1, -1, -1, -1, -1, -1},
{10, 7, 6, 1, 7, 10, 1, 8, 7, 1, 0, 8, -1, -1, -1, -1},
{0, 3, 7, 0, 7, 10, 0, 10, 9, 6, 10, 7, -1, -1, -1, -1},
{7, 6, 10, 7, 10, 8, 8, 10, 9, -1, -1, -1, -1, -1, -1, -1},
{6, 8, 4, 11, 8, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 6, 11, 3, 0, 6, 0, 4, 6, -1, -1, -1, -1, -1, -1, -1},
{8, 6, 11, 8, 4, 6, 9, 0, 1, -1, -1, -1, -1, -1, -1, -1},
{9, 4, 6, 9, 6, 3, 9, 3, 1, 11, 3, 6, -1, -1, -1, -1},
{6, 8, 4, 6, 11, 8, 2, 10, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 3, 0, 11, 0, 6, 11, 0, 4, 6, -1, -1, -1, -1},
{4, 11, 8, 4, 6, 11, 0, 2, 9, 2, 10, 9, -1, -1, -1, -1},
{10, 9, 3, 10, 3, 2, 9, 4, 3, 11, 3, 6, 4, 6, 3, -1},
{8, 2, 3, 8, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1},
{0, 4, 2, 4, 6, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 9, 0, 2, 3, 4, 2, 4, 6, 4, 3, 8, -1, -1, -1, -1},
{1, 9, 4, 1, 4, 2, 2, 4, 6, -1, -1, -1, -1, -1, -1, -1},
{8, 1, 3, 8, 6, 1, 8, 4, 6, 6, 10, 1, -1, -1, -1, -1},
{10, 1, 0, 10, 0, 6, 6, 0, 4, -1, -1, -1, -1, -1, -1, -1},
{4, 6, 3, 4, 3, 8, 6, 10, 3, 0, 3, 9, 10, 9, 3, -1},
{10, 9, 4, 6, 10, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 5, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 4, 9, 5, 11, 7, 6, -1, -1, -1, -1, -1, -1, -1},
{5, 0, 1, 5, 4, 0, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
{11, 7, 6, 8, 3, 4, 3, 5, 4, 3, 1, 5, -1, -1, -1, -1},
{9, 5, 4, 10, 1, 2, 7, 6, 11, -1, -1, -1, -1, -1, -1, -1},
{6, 11, 7, 1, 2, 10, 0, 8, 3, 4, 9, 5, -1, -1, -1, -1},
{7, 6, 11, 5, 4, 10, 4, 2, 10, 4, 0, 2, -1, -1, -1, -1},
{3, 4, 8, 3, 5, 4, 3, 2, 5, 10, 5, 2, 11, 7, 6, -1},
{7, 2, 3, 7, 6, 2, 5, 4, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 5, 4, 0, 8, 6, 0, 6, 2, 6, 8, 7, -1, -1, -1, -1},
{3, 6, 2, 3, 7, 6, 1, 5, 0, 5, 4, 0, -1, -1, -1, -1},
{6, 2, 8, 6, 8, 7, 2, 1, 8, 4, 8, 5, 1, 5, 8, -1},
{9, 5, 4, 10, 1, 6, 1, 7, 6, 1, 3, 7, -1, -1, -1, -1},
{1, 6, 10, 1, 7, 6, 1, 0, 7, 8, 7, 0, 9, 5, 4, -1},
{4, 0, 10, 4, 10, 5, 0, 3, 10, 6, 10, 7, 3, 7, 10, -1},
{7, 6, 10, 7, 10, 8, 5, 4, 10, 4, 8, 10, -1, -1, -1, -1},
{6, 9, 5, 6, 11, 9, 11, 8, 9, -1, -1, -1, -1, -1, -1, -1},
{3, 6, 11, 0, 6, 3, 0, 5, 6, 0, 9, 5, -1, -1, -1, -1},
{0, 11, 8, 0, 5, 11, 0, 1, 5, 5, 6, 11, -1, -1, -1, -1},
{6, 11, 3, 6, 3, 5, 5, 3, 1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 10, 9, 5, 11, 9, 11, 8, 11, 5, 6, -1, -1, -1, -1},
{0, 11, 3, 0, 6, 11, 0, 9, 6, 5, 6, 9, 1, 2, 10, -1},
{11, 8, 5, 11, 5, 6, 8, 0, 5, 10, 5, 2, 0, 2, 5, -1},
{6, 11, 3, 6, 3, 5, 2, 10, 3, 10, 5, 3, -1, -1, -1, -1},
{5, 8, 9, 5, 2, 8, 5, 6, 2, 3, 8, 2, -1, -1, -1, -1},
{9, 5, 6, 9, 6, 0, 0, 6, 2, -1, -1, -1, -1, -1, -1, -1},
{1, 5, 8, 1, 8, 0, 5, 6, 8, 3, 8, 2, 6, 2, 8, -1},
{1, 5, 6, 2, 1, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 3, 6, 1, 6, 10, 3, 8, 6, 5, 6, 9, 8, 9, 6, -1},
{10, 1, 0, 10, 0, 6, 9, 5, 0, 5, 6, 0, -1, -1, -1, -1},
{0, 3, 8, 5, 6, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{10, 5, 6, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 5, 10, 7, 5, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{11, 5, 10, 11, 7, 5, 8, 3, 0, -1, -1, -1, -1, -1, -1, -1},
{5, 11, 7, 5, 10, 11, 1, 9, 0, -1, -1, -1, -1, -1, -1, -1},
{10, 7, 5, 10, 11, 7, 9, 8, 1, 8, 3, 1, -1, -1, -1, -1},
{11, 1, 2, 11, 7, 1, 7, 5, 1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 1, 2, 7, 1, 7, 5, 7, 2, 11, -1, -1, -1, -1},
{9, 7, 5, 9, 2, 7, 9, 0, 2, 2, 11, 7, -1, -1, -1, -1},
{7, 5, 2, 7, 2, 11, 5, 9, 2, 3, 2, 8, 9, 8, 2, -1},
{2, 5, 10, 2, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1},
{8, 2, 0, 8, 5, 2, 8, 7, 5, 10, 2, 5, -1, -1, -1, -1},
{9, 0, 1, 5, 10, 3, 5, 3, 7, 3, 10, 2, -1, -1, -1, -1},
{9, 8, 2, 9, 2, 1, 8, 7, 2, 10, 2, 5, 7, 5, 2, -1},
{1, 3, 5, 3, 7, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 7, 0, 7, 1, 1, 7, 5, -1, -1, -1, -1, -1, -1, -1},
{9, 0, 3, 9, 3, 5, 5, 3, 7, -1, -1, -1, -1, -1, -1, -1},
{9, 8, 7, 5, 9, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{5, 8, 4, 5, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1},
{5, 0, 4, 5, 11, 0, 5, 10, 11, 11, 3, 0, -1, -1, -1, -1},
{0, 1, 9, 8, 4, 10, 8, 10, 11, 10, 4, 5, -1, -1, -1, -1},
{10, 11, 4, 10, 4, 5, 11, 3, 4, 9, 4, 1, 3, 1, 4, -1},
{2, 5, 1, 2, 8, 5, 2, 11, 8, 4, 5, 8, -1, -1, -1, -1},
{0, 4, 11, 0, 11, 3, 4, 5, 11, 2, 11, 1, 5, 1, 11, -1},
{0, 2, 5, 0, 5, 9, 2, 11, 5, 4, 5, 8, 11, 8, 5, -1},
{9, 4, 5, 2, 11, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 5, 10, 3, 5, 2, 3, 4, 5, 3, 8, 4, -1, -1, -1, -1},
{5, 10, 2, 5, 2, 4, 4, 2, 0, -1, -1, -1, -1, -1, -1, -1},
{3, 10, 2, 3, 5, 10, 3, 8, 5, 4, 5, 8, 0, 1, 9, -1},
{5, 10, 2, 5, 2, 4, 1, 9, 2, 9, 4, 2, -1, -1, -1, -1},
{8, 4, 5, 8, 5, 3, 3, 5, 1, -1, -1, -1, -1, -1, -1, -1},
{0, 4, 5, 1, 0, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{8, 4, 5, 8, 5, 3, 9, 0, 5, 0, 3, 5, -1, -1, -1, -1},
{9, 4, 5, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 11, 7, 4, 9, 11, 9, 10, 11, -1, -1, -1, -1, -1, -1, -1},
{0, 8, 3, 4, 9, 7, 9, 11, 7, 9, 10, 11, -1, -1, -1, -1},
{1, 10, 11, 1, 11, 4, 1, 4, 0, 7, 4, 11, -1, -1, -1, -1},
{3, 1, 4, 3, 4, 8, 1, 10, 4, 7, 4, 11, 10, 11, 4, -1},
{4, 11, 7, 9, 11, 4, 9, 2, 11, 9, 1, 2, -1, -1, -1, -1},
{9, 7, 4, 9, 11, 7, 9, 1, 11, 2, 11, 1, 0, 8, 3, -1},
{11, 7, 4, 11, 4, 2, 2, 4, 0, -1, -1, -1, -1, -1, -1, -1},
{11, 7, 4, 11, 4, 2, 8, 3, 4, 3, 2, 4, -1, -1, -1, -1},
{2, 9, 10, 2, 7, 9, 2, 3, 7, 7, 4, 9, -1, -1, -1, -1},
{9, 10, 7, 9, 7, 4, 10, 2, 7, 8, 7, 0, 2, 0, 7, -1},
{3, 7, 10, 3, 10, 2, 7, 4, 10, 1, 10, 0, 4, 0, 10, -1},
{1, 10, 2, 8, 7, 4, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 1, 4, 1, 7, 7, 1, 3, -1, -1, -1, -1, -1, -1, -1},
{4, 9, 1, 4, 1, 7, 0, 8, 1, 8, 7, 1, -1, -1, -1, -1},
{4, 0, 3, 7, 4, 3, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{4, 8, 7, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{9, 10, 8, 10, 11, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 9, 3, 9, 11, 11, 9, 10, -1, -1, -1, -1, -1, -1, -1},
{0, 1, 10, 0, 10, 8, 8, 10, 11, -1, -1, -1, -1, -1, -1, -1},
{3, 1, 10, 11, 3, 10, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 2, 11, 1, 11, 9, 9, 11, 8, -1, -1, -1, -1, -1, -1, -1},
{3, 0, 9, 3, 9, 11, 1, 2, 9, 2, 11, 9, -1, -1, -1, -1},
{0, 2, 11, 8, 0, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{3, 2, 11, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 8, 2, 8, 10, 10, 8, 9, -1, -1, -1, -1, -1, -1, -1},
{9, 10, 2, 0, 9, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{2, 3, 8, 2, 8, 10, 0, 1, 8, 1, 10, 8, -1, -1, -1, -1},
{1, 10, 2, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{1, 3, 8, 9, 1, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 9, 1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{0, 3, 8, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1},
{-1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1, -1}};

bool polygonizeCube(Cube cube, void(*callback)(Vector3,Vector3,Vector3))
{
   // Determine the index into the edge table which 
   // tells us which vertices are inside of the surface 
   int vertexFlags = 0;
   for (int i=0; i<8; i++)
	   if (cube.value[i] < 0) 
		   vertexFlags |= (1 << i);

	// Create the triangles 
	for (i=0; triTable[vertexFlags][i]!=-1; i+=3) 
	{
		Vector3 v0 = cube.intersections[triTable[vertexFlags][i  ]];
		Vector3 v1 = cube.intersections[triTable[vertexFlags][i+1]];
		Vector3 v2 = cube.intersections[triTable[vertexFlags][i+2]];
		callback(v0,v1,v2);
	}

	return true;
}

//#else

void polygonizeTetrahedron(Cube cube, Vector3 points[], int p0, int p1, int p2, int p3, void(*callback)(Vector3,Vector3,Vector3))
{
   // Determine which of the 16 cases we have given which vertices
   // are above or below the isosurface
   int vertexFlags = 0;
   if (cube.value[p0] < 0) vertexFlags |= 1;
   if (cube.value[p1] < 0) vertexFlags |= 2;
   if (cube.value[p2] < 0) vertexFlags |= 4;
   if (cube.value[p3] < 0) vertexFlags |= 8;

   Vector3 v0,v1,v2;
   switch (vertexFlags) 
   {
	   case 0:
	   case 15:
		  return;
	   case 1:
	   case 14:
		  v0 = interpolate(points[p0],points[p1],cube.value[p0],cube.value[p1]);
		  v1 = interpolate(points[p0],points[p2],cube.value[p0],cube.value[p2]);
		  v2 = interpolate(points[p0],points[p3],cube.value[p0],cube.value[p3]);
		  callback(v0,v1,v2);
		  return;
	   case 2:
	   case 13:
		  v0 = interpolate(points[p1],points[p0],cube.value[p1],cube.value[p0]);
		  v1 = interpolate(points[p1],points[p3],cube.value[p1],cube.value[p3]);
		  v2 = interpolate(points[p1],points[p2],cube.value[p1],cube.value[p2]);
		  callback(v0,v1,v2);
		  return;
	   case 3:
	   case 12:
		  v0 = interpolate(points[p0],points[p3],cube.value[p0],cube.value[p3]);
		  v1 = interpolate(points[p0],points[p2],cube.value[p0],cube.value[p2]);
		  v2 = interpolate(points[p1],points[p2],cube.value[p1],cube.value[p2]);
		  callback(v0,v1,v2);
		  v1 = interpolate(points[p1],points[p3],cube.value[p1],cube.value[p3]);
		  callback(v0,v1,v2);
		  return;
	   case 4:
	   case 11:
		  v0 = interpolate(points[p2],points[p0],cube.value[p2],cube.value[p0]);
		  v1 = interpolate(points[p2],points[p1],cube.value[p2],cube.value[p1]);
		  v2 = interpolate(points[p2],points[p3],cube.value[p2],cube.value[p3]);
		  callback(v0,v1,v2);
		  return;
	   case 5:
	   case 10:
		  v0 = interpolate(points[p0],points[p3],cube.value[p0],cube.value[p3]);
		  v1 = interpolate(points[p0],points[p1],cube.value[p0],cube.value[p1]);
		  v2 = interpolate(points[p2],points[p1],cube.value[p2],cube.value[p1]);
		  callback(v0,v1,v2);
		  v1 = interpolate(points[p2],points[p3],cube.value[p2],cube.value[p3]);
		  callback(v0,v1,v2);
		  return;
	   case 6:
	   case 9:
		  v0 = interpolate(points[p1],points[p3],cube.value[p1],cube.value[p3]);
		  v1 = interpolate(points[p1],points[p0],cube.value[p1],cube.value[p0]);
		  v2 = interpolate(points[p2],points[p0],cube.value[p2],cube.value[p0]);
		  callback(v0,v1,v2);
		  v1 = interpolate(points[p2],points[p3],cube.value[p2],cube.value[p3]);
		  callback(v0,v1,v2);
		  return;
	   case 7:
	   case 8:
		  v0 = interpolate(points[p3],points[p0],cube.value[p3],cube.value[p0]);
		  v1 = interpolate(points[p3],points[p2],cube.value[p3],cube.value[p2]);
		  v2 = interpolate(points[p3],points[p1],cube.value[p3],cube.value[p1]);
		  callback(v0,v1,v2);
		  return;
   }
}

bool polygonizeTetrahedra(Cube cube, void(*callback)(Vector3, Vector3, Vector3))
{
   // Determine the index into the edge table which 
   // tells us which vertices are inside of the surface 
   int vertexFlags = 0;
   for (int i=0; i<8; i++)
	   if (cube.value[i] < 0) 
		   vertexFlags |= (1 << i);

   // Cube is entirely in/out of the surface 
   if (cases[vertexFlags].type == 0)
      return false;

   Vector3 points[8];
	points[0] = Vector3(cube.min.x,cube.min.y,cube.min.z);
	points[1] = Vector3(cube.max.x,cube.min.y,cube.min.z);
	points[2] = Vector3(cube.max.x,cube.min.y,cube.max.z);
	points[3] = Vector3(cube.min.x,cube.min.y,cube.max.z);
	points[4] = Vector3(cube.min.x,cube.max.y,cube.min.z);
	points[5] = Vector3(cube.max.x,cube.max.y,cube.min.z);
	points[6] = Vector3(cube.max.x,cube.max.y,cube.max.z);
	points[7] = Vector3(cube.min.x,cube.max.y,cube.max.z);

   polygonizeTetrahedron(cube, points, 0,6,2,3, callback);
   polygonizeTetrahedron(cube, points, 0,6,1,2, callback);
   polygonizeTetrahedron(cube, points, 0,6,1,5, callback);
   polygonizeTetrahedron(cube, points, 0,3,6,7, callback);
   polygonizeTetrahedron(cube, points, 0,4,6,7, callback);
   polygonizeTetrahedron(cube, points, 0,6,4,5, callback);

   return true;
}

//#endif

bool polygonize(Cube cube, void(*callback)(Vector3, Vector3, Vector3))
{
   // Determine the index into the edge table which 
   // tells us which vertices are inside of the surface 
   int vertexFlags = 0;
   for (int i=0; i<8; i++)
	   if (cube.value[i] < 0) 
		   vertexFlags |= (1 << i);

   // Cube is entirely in/out of the surface 
   if (cases[vertexFlags].type == 0)
      return false;

   if (cases[vertexFlags].type == 1 ||
       cases[vertexFlags].type == 2 ||
       cases[vertexFlags].type == 5 ||
       cases[vertexFlags].type == 8 ||
       cases[vertexFlags].type == 9 ||
       cases[vertexFlags].type == 11 ||
       cases[vertexFlags].type == 14)
	   polygonizeCube(cube, callback);
   else
	   polygonizeTetrahedra(cube, callback);

   return true;
}


// generation of the topological-cases-table

Case simpleCases[15];
Case cases[256];

int makeVertexFlags(int p0, int p1, int p2, int p3, int p4, int p5, int p6, int p7)
{
	return p0 | (p1<<1) | (p2<<2) | (p3<<3) | (p4<<4) | (p5<<5) | (p6<<6) | (p7<<7);
}

int makeEdgeFlags(int p0, int p1, int p2, int p3, int p4, int p5, int p6, int p7, int p8,
				  int p9, int p10, int p11)
{
	return p0 | (p1<<1) | (p2<<2) | (p3<<3) | (p4<<4) | (p5<<5) | (p6<<6) | (p7<<7) 
		      | (p8<<8) | (p9<<9) | (p10<<10) | (p11<<11);
}

int getBit(int flag, int bit)
{
	return (flag >> bit) & 1;
}

void initializeSimpleCases()
{
	simpleCases[ 0].type = 0;
	simpleCases[ 0].vertexFlags = makeVertexFlags(0,0,0,0,0,0,0,0);
	simpleCases[ 0].edgeFlags = makeEdgeFlags(0,0,0,0,0,0,0,0,0,0,0,0);

	simpleCases[ 1].type = 1;
	simpleCases[ 1].vertexFlags = makeVertexFlags(0,0,0,0,0,0,1,0);
	simpleCases[ 1].edgeFlags = makeEdgeFlags(0,0,0,0,0,1,1,0,0,0,1,0);

	simpleCases[ 2].type = 2;
	simpleCases[ 2].vertexFlags = makeVertexFlags(0,0,0,1,0,0,0,1);
	simpleCases[ 2].edgeFlags = makeEdgeFlags(0,0,1,1,0,0,1,1,0,0,0,0);

	simpleCases[ 3].type = 3;
	simpleCases[ 3].vertexFlags = makeVertexFlags(0,0,0,1,0,0,1,0);
	simpleCases[ 3].edgeFlags = makeEdgeFlags(0,0,1,1,0,1,1,0,0,0,1,1);

	simpleCases[ 4].type = 4;
	simpleCases[ 4].vertexFlags = makeVertexFlags(0,0,0,1,0,1,0,0);
	simpleCases[ 4].edgeFlags = makeEdgeFlags(0,0,1,1,1,1,0,0,0,1,0,1);

	simpleCases[ 5].type = 5;
	simpleCases[ 5].vertexFlags = makeVertexFlags(1,1,1,0,0,0,0,0);
	simpleCases[ 5].edgeFlags = makeEdgeFlags(0,0,1,1,0,0,0,0,1,1,1,0);

	simpleCases[ 6].type = 6;
	simpleCases[ 6].vertexFlags = makeVertexFlags(0,0,0,1,0,1,0,1);
	simpleCases[ 6].edgeFlags = makeEdgeFlags(0,0,1,1,1,1,1,1,0,1,0,0);

	simpleCases[ 7].type = 7;
	simpleCases[ 7].vertexFlags = makeVertexFlags(0,0,1,0,0,1,0,1);
	simpleCases[ 7].edgeFlags = makeEdgeFlags(0,1,1,0,1,1,1,1,0,1,1,1);

	simpleCases[ 8].type = 8;
	simpleCases[ 8].vertexFlags = makeVertexFlags(1,1,1,1,0,0,0,0);
	simpleCases[ 8].edgeFlags = makeEdgeFlags(0,0,0,0,0,0,0,0,1,1,1,1);

	simpleCases[ 9].type = 9;
	simpleCases[ 9].vertexFlags = makeVertexFlags(1,1,0,1,1,0,0,0);
	simpleCases[ 9].edgeFlags = makeEdgeFlags(0,1,1,0,1,0,0,1,0,1,0,1);

	simpleCases[10].type = 10;
	simpleCases[10].vertexFlags = makeVertexFlags(0,1,0,1,0,1,0,1);
	simpleCases[10].edgeFlags = makeEdgeFlags(1,1,1,1,1,1,1,1,0,0,0,0);

	simpleCases[11].type = 11;
	simpleCases[11].vertexFlags = makeVertexFlags(1,1,0,1,0,1,0,0);
	simpleCases[11].edgeFlags = makeEdgeFlags(0,1,1,0,1,1,0,0,1,0,0,1);

	simpleCases[12].type = 12;
	simpleCases[12].vertexFlags = makeVertexFlags(1,1,1,0,0,0,0,1);
	simpleCases[12].edgeFlags = makeEdgeFlags(0,0,1,1,0,0,1,1,1,1,1,1);

	simpleCases[13].type = 13;
	simpleCases[13].vertexFlags = makeVertexFlags(0,1,0,1,1,0,1,0);
	simpleCases[13].edgeFlags = makeEdgeFlags(1,1,1,1,1,1,1,1,1,1,1,1);

	simpleCases[14].type = 14;
	simpleCases[14].vertexFlags = makeVertexFlags(1,1,1,0,1,0,0,0);
	simpleCases[14].edgeFlags = makeEdgeFlags(0,0,1,1,1,0,0,1,0,1,1,0);
}

int permuteVertexFlags(int flags, int v1, int v2, int v3, int v4, int v5, int v6, int v7, int v8)
{
	return makeVertexFlags( getBit(flags, v1),
						    getBit(flags, v2),
							getBit(flags, v3),
							getBit(flags, v4),
							getBit(flags, v5),
							getBit(flags, v6),
							getBit(flags, v7),
							getBit(flags, v8) );
}

int permuteEdgeFlags(int flags, int e1, int e2, int e3, int e4, int e5, int e6, int e7, int e8, int e9, int e10, int e11, int e12)
{
	return makeEdgeFlags(   getBit(flags, e1),
						    getBit(flags, e2),
							getBit(flags, e3),
							getBit(flags, e4),
							getBit(flags, e5),
							getBit(flags, e6),
							getBit(flags, e7),
							getBit(flags, e8),
							getBit(flags, e9),
							getBit(flags, e10),
							getBit(flags, e11),
							getBit(flags, e12) );
}

Case rotateX(int times, Case c)
{
	for (int i=0; i<times; i++)
	{
		c.vertexFlags = permuteVertexFlags(c.vertexFlags, 4,5,1,0,7,6,2,3);
		c.edgeFlags = permuteEdgeFlags(c.edgeFlags, 4,9,0,8,6,10,2,11,7,5,1,3);
	}
	return c;
}

Case rotateY(int times, Case c)
{
	for (int i=0; i<times; i++)
	{
		c.vertexFlags = permuteVertexFlags(c.vertexFlags, 1,2,3,0,5,6,7,4);
		c.edgeFlags = permuteEdgeFlags(c.edgeFlags, 1,2,3,0,5,6,7,4,9,10,11,8);
	}
	return c;
}

Case rotateZ(int times, Case c)
{
	for (int i=0; i<times; i++)
	{
		c.vertexFlags = permuteVertexFlags(c.vertexFlags, 1,5,6,2,0,4,7,3);
		c.edgeFlags = permuteEdgeFlags(c.edgeFlags, 9,5,10,1,8,7,11,3,0,4,6,2);
	}
	return c;
}

Case invert(int times, Case c)
{
	for (int i=0; i<times; i++)
		c.vertexFlags = ~c.vertexFlags;
	return c;
}

// generate the edgeTable
void initializePolygonisation()
{
	printf("topologically classifying cases ");

	initializeSimpleCases();

	for (int n=0; n<15; n++)
	{
		printf(".");

		Case simpleCase = simpleCases[n];

		for (int x=0; x<4; x++)
			for (int y=0; y<4; y++)
				for (int z=0; z<3; z++)
					for (int inv=0; inv<2; inv++)
					{
						Case c = rotateX(x,   rotateY(y,   rotateZ(z, 
								 invert(inv, simpleCase))));

						cases[c.vertexFlags] = c;
						cases[c.vertexFlags].simple = 
							simpleCase.type == 0 ||
							simpleCase.type == 1 ||
							simpleCase.type == 2 ||
							simpleCase.type == 5 ||
							simpleCase.type == 8 ||
							simpleCase.type == 9 ||
							simpleCase.type == 11 ||
							simpleCase.type == 14;
							
					}
	}

	printf(" ok\n");
}


